"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
function parse(input) {
    let i = 0;
    const impl = () => {
        while (input[i].match(/\s/))
            i++; // skip whitespaces
        if (input[i] === "(") {
            // drop '('
            i++;
        }
        else {
            throw new Error(`Input is not valid: unexpected '${input[i]}' at the beginning`);
        }
        const result = [];
        let node = "";
        while (true) {
            let c = input[i++];
            if (!c) {
                if (!node) {
                    break;
                }
                else {
                    throw new Error(`Input is not valid: unexpected '${node}' at the end`);
                }
            }
            if (c === ")") {
                if (node)
                    result.push(node);
                break;
            }
            else if (c === "\\") {
                c += input[i++];
                node += c;
            }
            else if (c.match(/\s/)) {
                if (node)
                    result.push(node);
                node = "";
            }
            else if (c === "(") {
                i--;
                result.push(impl());
            }
            else if (c === '"') {
                node += c;
                // parse string
                while ((c = input[i++])) {
                    // skip \"
                    if (c === "\\" && input[i] === '"') {
                        c += input[i++];
                    }
                    node += c;
                    if (c === '"')
                        break;
                }
            }
            else {
                node += c;
            }
        }
        return result;
    };
    return impl();
}
exports.parse = parse;
function length(sExp) {
    let sum = 0;
    for (const node of sExp) {
        sum += typeof node === "string" ? node.length : length(node);
    }
    return sum;
}
function indent(str) {
    return str.replace(/^/gm, "  ");
}
function stringify(node) {
    return typeof node === "string" ? node : beautify(node);
}
function beautify(sExp) {
    sExp = typeof sExp === "string" ? parse(sExp) : sExp;
    if (length(sExp) < 50) {
        return `(${sExp.map(stringify).join(" ")})`;
    }
    else {
        return (`(${sExp[0]}\n` +
            sExp
                .slice(1)
                .map(stringify)
                .map(indent)
                .join("\n") +
            "\n)");
    }
}
exports.beautify = beautify;
